@@ -1,8 +1,9 @@
-/* Licensed to the Apache Software Foundation (ASF) under one or more
- * contributor license agreements. 
- * The ASF licenses this file to You under the Apache License, Version 2.0
- * (the "License"); you may not use this file except in compliance with
- * the License.  You may obtain a copy of the License at
+/**
+ * Copyright 2006 Christian Liesch
+ *
+ * fooLicensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
  *
  *     http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -129,8 +130,12 @@ static apr_status_t global_EXEC(command_t *self, global_t *global,
 			       char *data, apr_pool_t *ptmp); 
 static apr_status_t global_SET(command_t *self, global_t *global, 
 			      char *data, apr_pool_t *ptmp); 
+static apr_status_t global_GLOBAL(command_t *self, global_t *global, 
+			          char *data, apr_pool_t *ptmp); 
+static apr_status_t global_PATH(command_t *self, global_t *global, 
+				char *data, apr_pool_t *ptmp); 
 static apr_status_t global_INCLUDE(command_t *self, global_t *global, 
-				  char *data, apr_pool_t *ptmp); 
+				   char *data, apr_pool_t *ptmp); 
 static apr_status_t global_TIMEOUT(command_t *self, global_t *global, 
 				  char *data, apr_pool_t *ptmp); 
 static apr_status_t global_AUTO_CLOSE(command_t *self, global_t *global, 
@@ -145,7 +150,7 @@ command_t global_commands[] = {
   "Close CLIENT|SERVER body",
   COMMAND_FLAGS_NONE},
   {"GO", (command_f )global_GO, "", 
-  "Starts all so far defined clients, servers and daemons",
+  "Starts all defined clients, servers and daemons. All started client and servers will be joined",
   COMMAND_FLAGS_NONE},
   {"CLIENT", (command_f )global_CLIENT, "[<number of concurrent clients>]", 
   "Client body start, close it with END and a newline",
@@ -154,7 +159,11 @@ command_t global_commands[] = {
   "Server body start, close it with END and a newline,\n"
   "Do load server.cert.pem and server.key.pem if found in local directory,\n"
   "number of concurrent servers, -1 for unlimited,\n"
-  "<SSL>: SSL, SSL2, SSL3, TLS1\n"
+  "<SSL>: SSL, SSL2, SSL3, DTLS1, TLS1"
+#if (OPENSSL_VERSION_NUMBER >= 0x1000102fL)
+  ", TLS1.1, TLS1.2"
+#endif
+  "\n"
   "<addr_port>: 8080                (just the port number)\n"
   "             www.apache.org      (just the hostname)\n"
   "             www.apache.org:8080 (hostname and port number)\n"
@@ -166,6 +175,12 @@ command_t global_commands[] = {
   {"SET", (command_f )global_SET, "<variable>=<value>", 
   "Store a value in a global variable",
   COMMAND_FLAGS_NONE},
+  {"GLOBAL", (command_f )global_GLOBAL, "<variable-name>+", 
+  "Define the given variable as global, this is shared over all threads",
+  COMMAND_FLAGS_NONE},
+  {"PATH", (command_f )global_PATH, "<include paths colon separated>", 
+  "Defines a set of path where INCLUDE looks first for there include files",
+  COMMAND_FLAGS_NONE},
   {"INCLUDE", (command_f )global_INCLUDE, "<include file>", 
   "Load and execute defined include file,\n"
   "current path is taken the callers current path",
@@ -219,17 +234,22 @@ command_t local_commands[] = {
   {"_REQ", (command_f )command_REQ, "<host> [<SSL>:]<port>[:<tag>] [<cert-file> <key-file> [<ca-cert-file>]]", 
   "Open connection to defined host:port, with SSL support.\n"
   "If connection exist no connect will be performed\n"
-  "<SSL>: SSL, SSL2, SSL3, TLS1\n"
+  "<SSL>: SSL, SSL2, SSL3, DTLS1, TLS1"
+#if (OPENSSL_VERSION_NUMBER >= 0x1000102fL)
+  ", TLS1.1, TLS1.2"
+#endif
+  "\n"
   "<host>: host name or IPv4/IPv6 address (IPv6 address must be surrounded\n"
   "        in square brackets)\n"
   "<tag>: Additional tag info do support multiple connection to one target\n"
   "<cert-file>, <key-file> and <ca-cert-file> are optional for client/server authentication",
   COMMAND_FLAGS_NONE},	
   {"_RESWAIT", (command_f )command_RESWAIT, "", 
-   "Combines the _RES and _WAIT command ignoring TCP connections not sending any data." ,
-  COMMAND_FLAGS_NONE},
-  {"_RES", (command_f )command_RES, "", 
-  "Wait for a connection accept",
+   "Do use _RES IGNORE_MONITORS instead" ,
+  COMMAND_FLAGS_DEPRECIATED},
+  {"_RES", (command_f )command_RES, "[IGNORE_MONITORS]", 
+  "Wait for a connection accept \n"
+  "IGNORE_MONITORS do ignore all connection pings without data",
   COMMAND_FLAGS_NONE},
   {"_WAIT", (command_f )command_WAIT, "[<amount of bytes>]", 
   "Wait for data and receive them.\n"
@@ -249,6 +269,9 @@ command_t local_commands[] = {
   {"_GREP", (command_f )command_GREP, "(.|headers|body|error|exec|var()) \"|'<regex>\"|' <variable>", 
    "Define a regex with a match which should be stored in <variable> and do not fail if no match",
   COMMAND_FLAGS_NONE},
+  {"_ASSERT", (command_f )command_ASSERT, "<expression>", 
+   "Check if expression is true fail otherwise",
+  COMMAND_FLAGS_NONE},
   {"_SEQUENCE", (command_f )command_MATCH_SEQ, "<var-sequence>", 
    "Define a sequence of _MATCH variables which must apear in this order",
   COMMAND_FLAGS_NONE},
@@ -352,7 +375,11 @@ command_t local_commands[] = {
   {"_TUNNEL", (command_f )command_TUNNEL, "<host> [<SSL>:]<port>[:<tag>] [<cert-file> <key-file> [<ca-cert-file>]]", 
   "Open tunnel to defined host:port, with SSL support.\n"
   "If connection exist no connect will be performed\n"
-  "<SSL>: SSL, SSL2, SSL3, TLS1\n"
+  "<SSL>: SSL, SSL2, SSL3, DTLS1, TLS1"
+#if (OPENSSL_VERSION_NUMBER >= 0x1000102fL)
+  ", TLS1.1, TLS1.2"
+#endif
+  "\n"
   "<tag>:Additional tag info do support multiple connection to one target\n"
   "<cert-file>, <key-file> and <ca-cert-file> are optional for client/server authentication",
   COMMAND_FLAGS_NONE},	
@@ -731,19 +758,8 @@ static apr_status_t worker_where_is_else(worker_t *worker, int *else_pos) {
 static apr_status_t command_IF(command_t * self, worker_t * worker,
                                char *data, apr_pool_t *ptmp) {
   char *copy;
-  char *left;
-  char *right;
-  apr_ssize_t left_val;
-  apr_ssize_t right_val;
-  char *middle;
-  char *last;
-  const char *err;
-  int off;
-  regex_t *compiled;
   apr_status_t status;
   worker_t *body;
-  apr_size_t len;
-
   int doit = 0;
   int not = 0;
   int else_pos = 0;
@@ -769,16 +785,28 @@ static apr_status_t command_IF(command_t * self, worker_t * worker,
     }
     doit = val;
   }
-  else if (copy[0] =='"') {
+  else {
     /* old behavour */
-    ++copy;
-    left = apr_strtok(copy, "\"", &last);
-    middle = apr_strtok(NULL, " ", &last);
+    char *left;
+    char *right;
+    apr_ssize_t left_val;
+    apr_ssize_t right_val;
+    char *middle;
+    const char *err;
+    int off;
+    regex_t *compiled;
+    apr_size_t len;
+    char **argv;
+    int i = 0;
+
+    my_tokenize_to_argv(copy, &argv, ptmp, 0);
+    left = argv[i]; i++;
+    middle = argv[i]; i++;
     if (strcmp(middle, "NOT") == 0) {
       not = 1;
-      middle = apr_strtok(NULL, " ", &last);
+      middle = argv[i]; i++;
     }
-    right = apr_strtok(NULL, "\"", &last);
+    right = argv[i];
    
     if (!left || !middle || !right) {
       worker_log(worker, LOG_ERR, "%s: Syntax error '%s'", self->name, data);
@@ -905,16 +933,22 @@ static apr_status_t command_LOOP(command_t *self, worker_t *worker,
   worker_t *body;
   char *copy;
   int loop;
+  char **argv;
   int i;
+  char *var;
 
   COMMAND_NEED_ARG("<number>|FOREVER"); 
  
-  if (strncmp(copy, "FOREVER", 7) == 0) {
+  my_tokenize_to_argv(copy, &argv, ptmp, 0);
+
+  if (strncmp(argv[0], "FOREVER", 7) == 0) {
     loop = -1;
   }
   else {
-    loop = apr_atoi64(copy);
+    loop = apr_atoi64(argv[0]);
   }
+
+  var = argv[1]; 
   
   /* create a new worker body */
   if ((status = worker_body(&body, worker)) != APR_SUCCESS) {
@@ -924,6 +958,9 @@ static apr_status_t command_LOOP(command_t *self, worker_t *worker,
   /* loop */
   for (i = 0; loop == -1 || i < loop; i++) {
     /* interpret */
+    if (var) {
+      worker_var_set(body, var, apr_itoa(ptmp, i));
+    }
     if ((status = body->interpret(body, worker, NULL)) != APR_SUCCESS) {
       break;
     }
@@ -962,12 +999,13 @@ static apr_status_t command_FOR(command_t *self, worker_t *worker,
   char *var;
   char *list;
   char *cur;
+  char **argv;
 
   COMMAND_NEED_ARG("<variable> \"<string>*\""); 
  
-  var = apr_strtok(copy, " ", &last);
-  
-  list = my_unescape(last, &last);
+  my_tokenize_to_argv(copy, &argv, ptmp, 0);
+  var = argv[0];
+  list = argv[1];
 
   /* create a new worker body */
   if ((status = worker_body(&body, worker)) != APR_SUCCESS) {
@@ -1010,21 +1048,19 @@ static apr_status_t command_BPS(command_t *self, worker_t *worker, char *data,
                                 apr_pool_t *ptmp) {
   apr_status_t status;
   worker_t *body;
-  char *last;
   char *copy;
-  char *val;
   int bps;
   int duration;
+  char **argv;
   apr_time_t init;
   apr_time_t start;
   apr_time_t cur;
 
   COMMAND_NEED_ARG("Byte/s and duration time in second"); 
 
-  val = apr_strtok(copy, " ", &last);
-  bps = apr_atoi64(val);
-  val = apr_strtok(NULL, " ", &last);
-  duration = apr_atoi64(val);
+  my_tokenize_to_argv(copy, &argv, ptmp, 0);
+  bps = apr_atoi64(argv[0]);
+  duration = apr_atoi64(argv[1]);
   
   /* create a new worker body */
   if ((status = worker_body(&body, worker)) != APR_SUCCESS) {
@@ -1085,9 +1121,8 @@ static apr_status_t command_RPS(command_t *self, worker_t *worker, char *data,
                                 apr_pool_t *ptmp) {
   apr_status_t status;
   worker_t *body;
-  char *last;
   char *copy;
-  char *val;
+  char **argv;
   int rps;
   int duration;
   apr_time_t init;
@@ -1096,10 +1131,9 @@ static apr_status_t command_RPS(command_t *self, worker_t *worker, char *data,
 
   COMMAND_NEED_ARG("Byte/s and duration time in second"); 
 
-  val = apr_strtok(copy, " ", &last);
-  rps = apr_atoi64(val);
-  val = apr_strtok(NULL, " ", &last);
-  duration = apr_atoi64(val);
+  my_tokenize_to_argv(copy, &argv, ptmp, 0);
+  rps = apr_atoi64(argv[0]);
+  duration = apr_atoi64(argv[1]);
   
   /* create a new worker body */
   if ((status = worker_body(&body, worker)) != APR_SUCCESS) {
@@ -1169,7 +1203,7 @@ static apr_status_t command_ERROR(command_t *self, worker_t *worker,
 
   COMMAND_NEED_ARG("<error>"); 
  
- if ((status = apr_tokenize_to_argv(copy, &argv, ptmp)) == APR_SUCCESS) {
+ if ((status = my_tokenize_to_argv(copy, &argv, ptmp, 0)) == APR_SUCCESS) {
     if (!argv[0]) {
       worker_log_error(worker, "No argument found, need an regex for expected errof.");
       return APR_EINVAL;
@@ -1206,6 +1240,9 @@ static apr_status_t command_ERROR(command_t *self, worker_t *worker,
 
   worker_log(worker, LOG_CMD, "_END ERROR");
   
+  if (worker->socket) {
+    worker->socket->config = apr_hash_make(worker->pbody);
+  }
   worker_body_end(body, worker);
   return status;
 }
@@ -1355,6 +1392,8 @@ static apr_status_t worker_interpret(worker_t * worker, worker_t *parent,
     line = e[worker->cmd].val;
     if (worker_is_block(worker, line, ptmp)) {
       status = command_CALL(NULL, worker, line, ptmp);
+      /* dirty hack */
+      worker->file_and_line = e[worker->cmd].key;
       status = worker_check_error(parent, status);
     }
     else {
@@ -1456,6 +1495,7 @@ void worker_finally(worker_t *worker, apr_status_t status) {
     exit(1);
   }
 exodus:
+  htt_run_worker_finally(worker); 
   worker_conn_close_all(worker);
   apr_thread_exit(worker->mythread, APR_SUCCESS);
 }
@@ -1853,6 +1893,22 @@ static apr_status_t global_new(global_t **global, store_t *vars,
   return APR_SUCCESS;
 }
 
+/**
+ * cleanup files on exit
+ *
+ * @param data IN file name to remove
+ * @return APR_SUCCESS
+ */
+static apr_status_t worker_file_cleanup(void *data) {
+  const char *name = data;
+  apr_pool_t *pool;
+
+  apr_pool_create(&pool, NULL);
+  apr_file_remove(name, pool);
+  apr_pool_destroy(pool);
+  return APR_SUCCESS;
+}
+
 /**
  * Global CLIENT command
  *
@@ -1927,8 +1983,9 @@ static apr_status_t global_END(command_t *self, global_t *global, char *data,
 	      my_status_str(global->pool, status), status);
       return status;
     }
-    apr_table_addn(global->files, global->worker->name, 
-	           (const char *)global->worker);
+
+    apr_pool_cleanup_register(global->pool, global->worker->name, 
+                              worker_file_cleanup, apr_pool_cleanup_null);
     global->state = GLOBAL_STATE_NONE;
     return APR_SUCCESS;
     break; 
@@ -1992,8 +2049,6 @@ static apr_status_t global_worker(command_t *self, global_t *global, char *data,
     fprintf(stderr, "\nGlobal worker: could not create worker");
     return status;
   }
-  global->prefix = apr_pstrcat(global->pool, global->prefix, 
-			     "                        ", NULL);
   return APR_SUCCESS;
 }
 
@@ -2008,7 +2063,11 @@ static apr_status_t global_worker(command_t *self, global_t *global, char *data,
  */
 static apr_status_t global_CLIENT(command_t *self, global_t *global, char *data, 
                                   apr_pool_t *ptmp) {
-  return global_worker(self, global, data, GLOBAL_STATE_CLIENT);
+  apr_status_t status;
+  status = global_worker(self, global, data, GLOBAL_STATE_CLIENT);
+  global->prefix = apr_pstrcat(global->pool, global->prefix, 
+			       "                        ", NULL);
+  return status;
 }
 
 /**
@@ -2022,7 +2081,12 @@ static apr_status_t global_CLIENT(command_t *self, global_t *global, char *data,
  */
 static apr_status_t global_SERVER(command_t *self, global_t *global, char *data, 
                                   apr_pool_t *ptmp) {
-  return global_worker(self, global, data, GLOBAL_STATE_SERVER);
+  apr_status_t status;
+  status = global_worker(self, global, data, GLOBAL_STATE_SERVER);
+  global->prefix = apr_pstrcat(global->pool, global->prefix, 
+			       "                        ", NULL);
+
+  return status;
 }
 
 /**
@@ -2060,9 +2124,6 @@ static apr_status_t global_BLOCK(command_t * self, global_t * global,
     return status;
   }
   
-  /* A block has its callies prefix I suppose */
-  global->prefix = apr_pstrcat(global->pool, global->prefix, "", NULL);
-  
   /* Get params and returns */
   /* create two tables for in/out vars */
   /* input and output vars */
@@ -2122,9 +2183,6 @@ static apr_status_t global_FILE(command_t * self, global_t * global,
   }
 
   global->worker->name = data;
-  
-  /* A block has its callies prefix I suppose */
-  global->prefix = apr_pstrcat(global->pool, global->prefix, "", NULL);
 
   /* open file */
   return APR_SUCCESS;
@@ -2221,6 +2279,44 @@ static apr_status_t global_SET(command_t *self, global_t *global, char *data,
   return APR_SUCCESS;
 }
 
+/**
+ * Global GLOBAL command
+ *
+ * @param self IN command
+ * @param global IN global object
+ * @param data IN key=value
+ *
+ * @return APR_SUCCESS
+ */
+static apr_status_t global_GLOBAL(command_t *self, global_t *global, char *data, 
+                                  apr_pool_t *ptmp) {
+  char *last;
+  char *var;
+  
+  int i = 0;
+  
+  while (data[i] == ' ') ++i;
+
+  if (!global->shared) {
+    global->shared = store_make(global->pool);
+  }
+
+  var = apr_strtok(&data[i], " ", &last);
+  while (var) {
+    for (i = 0; var[i] != 0 && strchr(VAR_ALLOWED_CHARS, var[i]); i++); 
+    if (var[i] != 0) {
+      fprintf(stderr, "\nChar '%c' is not allowed in \"%s\"", var[i], var);
+      success = 0;
+      return APR_EINVAL;
+    }
+
+    store_set(global->shared, var, "");
+    var = apr_strtok(NULL, " ", &last);
+  }
+
+  return APR_SUCCESS;
+}
+
 /**
  * Use to define a MODULE. Used to make a name space for BLOCKs.
  *
@@ -2252,6 +2348,24 @@ static apr_status_t global_MODULE(command_t * self, global_t * global,
   return APR_SUCCESS;
 }
 
+/**
+ * Global PATH command
+ *
+ * @param self IN command
+ * @param global IN global object
+ * @param data IN path string
+ *
+ * @return APR_SUCCESS
+ */
+static apr_status_t global_PATH(command_t *self, global_t *global, char *data, 
+                                apr_pool_t *ptmp) {
+  char **argv;
+
+  my_tokenize_to_argv(data, &argv, global->pool, 0);
+  global->path = argv[0];
+  return APR_SUCCESS;
+}
+
 /**
  * Global INCLUDE command
  *
@@ -2259,7 +2373,7 @@ static apr_status_t global_MODULE(command_t * self, global_t * global,
  * @param global IN global object
  * @param data IN relative to caller or absolut path
  *
- * @return APR_SUCCESS
+ * @return APR_SUCCESS or APR_ENOENT if no include file found
  */
 static apr_status_t interpret_recursiv(apr_file_t *fp, global_t *global); 
 static apr_status_t global_INCLUDE(command_t *self, global_t *global, char *data, 
@@ -2271,13 +2385,28 @@ static apr_status_t global_INCLUDE(command_t *self, global_t *global, char *data
   int i;
 
   status = APR_ENOENT;
-  if (apr_tokenize_to_argv(data, &argv, global->pool) == APR_SUCCESS) {
-    for (i = 0; argv[i] != NULL; i++) {
-      /* open include file */
+  my_tokenize_to_argv(data, &argv, global->pool, 0);
+  for (i = 0; argv[i] != NULL; i++) {
+    if (argv[i][0] == '/' || global->path == NULL) {
       if ((status =
-	   apr_file_open(&fp, argv[i], APR_READ, APR_OS_DEFAULT,
-			 global->pool)) == APR_SUCCESS) {
-	break;
+           apr_file_open(&fp, argv[i], APR_READ, APR_OS_DEFAULT,
+                         global->pool)) == APR_SUCCESS) {
+        break;
+      }
+    }
+    else if (global->path) {
+      char *last;
+      char *cur;
+      char *path = apr_pstrdup(global->pool, global->path);
+
+      cur = apr_strtok(path, ":", &last);
+      while (cur) {
+        char *file = apr_pstrcat(global->pool, cur, "/", argv[i], NULL);
+        if ((status = apr_file_open(&fp, file, APR_READ, APR_OS_DEFAULT, 
+                                    global->pool)) == APR_SUCCESS) {
+          break;
+        }
+        cur = apr_strtok(NULL, ":", &last);
       }
     }
   }
@@ -2445,6 +2574,7 @@ static apr_status_t global_GO(command_t *self, global_t *global, char *data,
   worker_t *worker;
   apr_thread_t *thread;
 
+
   /* start all daemons first */
   e = (apr_table_entry_t *) apr_table_elts(global->daemons)->elts;
   for (i = 0; i < apr_table_elts(global->daemons)->nelts; ++i) {
@@ -2478,16 +2608,56 @@ static apr_status_t global_GO(command_t *self, global_t *global, char *data,
   e = (apr_table_entry_t *) apr_table_elts(global->clients)->elts;
   for (i = 0; i < apr_table_elts(global->clients)->nelts; ++i) {
     worker = (void *)e[i].val;
-    if ((status =
-	 apr_thread_create(&thread, global->tattr, worker_thread_client,
-			   worker, global->pool)) != APR_SUCCESS) {
-      fprintf(stderr, "\nCould not create client thread");
+    status = htt_run_client_create(worker, worker_thread_client, &thread);
+    if (status == APR_ENOTHREAD || status == APR_ENOTIMPL) {
+      if ((status =
+           apr_thread_create(&thread, global->tattr, worker_thread_client,
+                             worker, global->pool)) != APR_SUCCESS) {
+        fprintf(stderr, "\nCould not create client thread");
+        return status;
+      }
+    }
+    else if (status != APR_SUCCESS) {
       return status;
     }
-    apr_table_addn(global->threads, worker->name, (char *) thread);
+    if (thread) {
+      apr_table_addn(global->threads, worker->name, (char *) thread);
+    }
   }
   apr_table_clear(global->clients);
+   
+  /* start clients */
+  e = (apr_table_entry_t *) apr_table_elts(global->threads)->elts;
+  for (i = 0; i < apr_table_elts(global->threads)->nelts; ++i) {
+    thread = (apr_thread_t *) e[i].val;
+    status = htt_run_thread_start(global, thread);
+    if (status != APR_SUCCESS) {
+      fprintf(stderr, "\nCould not start client thread: %d", status);
+      return status;
+    }
+  }
+ 
+  /* wait on thermination of all started threads */
+  e = (apr_table_entry_t *) apr_table_elts(global->threads)->elts;
+  for (i = 0; i < apr_table_elts(global->threads)->nelts; ++i) {
+    apr_status_t retstat;
+    thread = (apr_thread_t *) e[i].val;
+    status = htt_run_thread_join(global, thread);
+    if (status == APR_ENOTHREAD || status == APR_ENOTIMPL) {
+      if ((retstat = apr_thread_join(&status, thread))) {
+        fprintf(stderr, "\nCould not join thread: %d", retstat);
+        return retstat;
+      }
+    }
+    else if (status != APR_SUCCESS) {
+      fprintf(stderr, "\nCould not join thread: %d", status);
+      return status;
+    }
+  }
+  apr_table_clear(global->threads);
 
+  htt_run_worker_joined(global);
+  global->prefix = apr_pstrdup(global->pool, "");
   return APR_SUCCESS;
 }
 
@@ -2597,12 +2767,8 @@ static apr_status_t interpret_recursiv(apr_file_t *fp, global_t *global) {
 static apr_status_t interpret(apr_file_t * fp, store_t * vars,
                               int log_mode, apr_pool_t * p, char *additional) {
   apr_status_t status;
-  apr_status_t retstat = APR_SUCCESS;
-  apr_table_entry_t *e;
   int i;
-  const char *name;
   global_t *global;
-  apr_thread_t *thread;
 
   if ((status = global_new(&global, vars, log_mode, p)) 
       != APR_SUCCESS) {
@@ -2639,24 +2805,9 @@ static apr_status_t interpret(apr_file_t * fp, store_t * vars,
     return status;
   }
 
-  global_GO(&global_commands[1], global, NULL, NULL);
-  
-  /* wait on thermination of all started threads */
-  e = (apr_table_entry_t *) apr_table_elts(global->threads)->elts;
-  for (i = 0; i < apr_table_elts(global->threads)->nelts; ++i) {
-    thread = (apr_thread_t *) e[i].val;
-    name = e[i].key;
-    if ((retstat = apr_thread_join(&status, thread))) {
-      fprintf(stderr, "\nCould not join thread: %d", retstat);
-      return retstat;
-    }
-    if (status != APR_SUCCESS) {
-      fprintf(stderr, "\nCould not join thread: %d", status);
-      return status;
-    }
-  }
+  status = global_GO(&global_commands[1], global, NULL, NULL);
 
-  return retstat;
+  return status;
 }
 
 apr_getopt_option_t options[] = {
@@ -2671,8 +2822,11 @@ apr_getopt_option_t options[] = {
   { "debug-system", 'p', 0, "log level debug-system to log more details" },
   { "list-commands", 'L', 0, "List all available script commands" },
   { "help-command", 'C', 1, "Print help for specific command" },
+  { "duration", 't', 0, "Print test duration" },
   { "timestamp", 'T', 0, "Time stamp on every run" },
   { "shell", 'S', 0, "Shell mode" },
+  { "shell", 'S', 0, "Shell mode" },
+  { "define", 'D', 1, "Define variables" },
   { NULL, 0, 0, NULL }
 };
 
@@ -2893,12 +3047,12 @@ static void show_command_help(apr_pool_t *p, global_t *global,
     block_name = apr_pstrcat(p, "_", last, NULL);
     if (!(blocks = apr_hash_get(global->modules, module, APR_HASH_KEY_STRING))) {
       fprintf(stdout, "\ncommand: %s do not exist\n\n", command);
-      goto exit;
+      exit(1);
     }
     block_name = apr_pstrcat(p, "_", last, NULL);
     if (!(worker = apr_hash_get(blocks, block_name, APR_HASH_KEY_STRING))) {
       fprintf(stdout, "\ncommand: %s do not exist\n\n", command);
-      goto exit;
+      exit(1);
     }
     else {
       char *help;
@@ -2917,6 +3071,7 @@ static void show_command_help(apr_pool_t *p, global_t *global,
   }
 
   fprintf(stdout, "\ncommand: %s do not exist\n\n", command);
+  exit(1);
 
 exit:
   fprintf(stdout, "\n");
@@ -2927,18 +3082,6 @@ exit:
  * own exit func
  */
 static void my_exit() {
-  int i;
-  worker_t *worker;
-
-  if (process_global) {
-    apr_table_entry_t *e = 
-      (apr_table_entry_t *) apr_table_elts(process_global->files)->elts;
-    for (i = 0; i < apr_table_elts(process_global->files)->nelts; i++) {
-      worker = (worker_t *)e[i].val;
-      apr_file_remove(worker->name, process_global->pool);
-    }
-  }
-
   if (!success) {
     fprintf(stderr, " FAILED\n");
     fflush(stderr);
@@ -2974,6 +3117,7 @@ int main(int argc, const char *const argv[]) {
 #define MAIN_FLAGS_PRINT_TSTAMP 1
 #define MAIN_FLAGS_USE_STDIN 2
 #define MAIN_FLAGS_NO_OUTPUT 4
+#define MAIN_FLAGS_PRINT_DURATION 8
   int flags;
   apr_time_t time;
   char time_str[256];
@@ -2992,6 +3136,9 @@ int main(int argc, const char *const argv[]) {
   log_mode = LOG_CMD;
   flags = MAIN_FLAGS_NONE;
 
+  /* create a global vars table */
+  vars = store_make(pool);
+
   /* get options */
   apr_getopt_init(&opt, pool, argc, argv);
   while ((status = apr_getopt_long(opt, options, &c, &optarg)) == APR_SUCCESS) {
@@ -3025,6 +3172,9 @@ int main(int argc, const char *const argv[]) {
     case 'd':
       log_mode = LOG_ALL_CMD;
       break;
+    case 't':
+      flags |= MAIN_FLAGS_PRINT_DURATION; 
+      break;
     case 'L':
       interpret(NULL, NULL, -1, pool, NULL);
       break;
@@ -3038,6 +3188,23 @@ int main(int argc, const char *const argv[]) {
     case 'S':
       flags |= MAIN_FLAGS_USE_STDIN; 
       break;
+    case 'D':
+      {
+        char *val;
+        char *var;
+        char *entry = apr_pstrdup(pool, optarg);
+
+        var = apr_strtok(entry, "=", &val);
+        if (val && val[0]) {
+          store_set(vars, var, val);
+        }
+        else {
+          fprintf(stderr, "Error miss value in variable definition \"-D%s\", need the format -D<var>=<val>\n", optarg);
+          fflush(stderr);
+          exit(1);
+        }
+      }
+      break;
     }
   }
 
@@ -3111,15 +3278,26 @@ int main(int argc, const char *const argv[]) {
       exit(1);
     }
 
-    /* create a global vars table */
-    vars = store_make(pool);
-
+    if (flags & MAIN_FLAGS_PRINT_DURATION) {
+      time = apr_time_now();
+    }
     /* interpret current file */
     if ((status = interpret(fp, vars, log_mode, pool, NULL)) != APR_SUCCESS) {
       success = 0;
       exit(1);
     }
 
+    if (log_mode > LOG_WARN) {
+      fprintf(stdout, "\n");
+      fflush(stdout);
+    }
+
+    if (flags & MAIN_FLAGS_PRINT_DURATION) {
+      time = apr_time_now() - time;
+      fprintf(stdout, "%"APR_TIME_T_FMT , time);
+      fflush(stdout);
+    }
+
     /* close current file */
     apr_file_close(fp);
 
@@ -3139,6 +3317,11 @@ APR_HOOK_STRUCT(
   APR_HOOK_LINK(block_end)
   APR_HOOK_LINK(server_port_args)
   APR_HOOK_LINK(worker_clone)
+  APR_HOOK_LINK(client_create)
+  APR_HOOK_LINK(thread_start)
+  APR_HOOK_LINK(worker_finally)
+  APR_HOOK_LINK(thread_join)
+  APR_HOOK_LINK(worker_joined)
 )
 
 APR_IMPLEMENT_EXTERNAL_HOOK_RUN_FIRST(htt, HTT, apr_status_t, server_port_args, 
@@ -3161,3 +3344,22 @@ APR_IMPLEMENT_EXTERNAL_HOOK_RUN_FIRST(htt, HTT, apr_status_t, block_end,
                                       (global_t *global), 
                                       (global), APR_SUCCESS);
 
+APR_IMPLEMENT_EXTERNAL_HOOK_RUN_FIRST(htt, HTT, apr_status_t, client_create, 
+                                      (worker_t *worker, apr_thread_start_t func, apr_thread_t **new_thread), 
+                                      (worker, func, new_thread), APR_ENOTIMPL);
+
+APR_IMPLEMENT_EXTERNAL_HOOK_RUN_FIRST(htt, HTT, apr_status_t, thread_start, 
+                                      (global_t *global, apr_thread_t *thread), 
+                                      (global, thread), APR_SUCCESS);
+
+APR_IMPLEMENT_EXTERNAL_HOOK_RUN_FIRST(htt, HTT, apr_status_t, worker_finally, 
+                                      (worker_t *worker), 
+                                      (worker), APR_SUCCESS);
+
+APR_IMPLEMENT_EXTERNAL_HOOK_RUN_FIRST(htt, HTT, apr_status_t, thread_join, 
+                                      (global_t *global, apr_thread_t *thread), 
+                                      (global, thread), APR_ENOTIMPL);
+
+APR_IMPLEMENT_EXTERNAL_HOOK_RUN_FIRST(htt, HTT, apr_status_t, worker_joined, 
+                                      (global_t *global), 
+                                      (global), APR_SUCCESS);
