@@ -1,8 +1,9 @@
-/* Licensed to the Apache Software Foundation (ASF) under one or more
- * contributor license agreements. 
- * The ASF licenses this file to You under the Apache License, Version 2.0
- * (the "License"); you may not use this file except in compliance with
- * the License.  You may obtain a copy of the License at
+/**
+ * Copyright 2010 Christian Liesch
+ *
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
  *
  *     http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -24,10 +25,14 @@
 /************************************************************************
  * Includes
  ***********************************************************************/
-#include <lua5.1/lua.h>
-#include <lua5.1/lualib.h>
-#include <lua5.1/lauxlib.h>
- 
+#define LUA_COMPAT_MODULE
+#include <lua.h>
+#include <lualib.h>
+#include <lauxlib.h>
+
+#include <apr_sha1.h>
+
+#include "lua_crypto.h"
 #include "module.h"
 
 /************************************************************************
@@ -142,21 +147,20 @@ static const char *lua_get_line(lua_State *L, void *ud, size_t *size) {
 /**
  * Do push the httest version on the stack
  * @lua_return version as a string
- * @param lua IN lua state
+ * @param L IN lua state
  * @return 1
  */
-static int lua_version(lua_State *lua) {
-  lua_pushstring(lua, PACKAGE_VERSION);
+static int luam_version(lua_State *L) {
+  lua_pushstring(L, PACKAGE_VERSION);
   return 1;
 }
 
 /**
  * Execute httest script.
- * @lua_arg string IN httest script
- * @param lua IN lua state
+ * @param L IN lua state
  * @return 0
  */
-static int lua_interpret(lua_State *lua) {
+static int luam_interpret(lua_State *L) {
   apr_status_t status;
   apr_pool_t *ptmp;
   worker_t *worker;
@@ -169,21 +173,21 @@ static int lua_interpret(lua_State *lua) {
   char *line;
   apr_size_t len;
 
-  if (!lua_isstring(lua, -1)) {
-    luaL_error(lua, "Expect a string to interpret");
+  if (!lua_isstring(L, -1)) {
+    luaL_error(L, "Expect a string to interpret");
     return 1;
   }
 
-  string = lua_tolstring(lua, -1, &len);
-  lua_pop(lua, 1);
+  string = lua_tolstring(L, -1, &len);
+  lua_pop(L, 1);
 
-  lua_getfield(lua, LUA_REGISTRYINDEX, "htt_worker");
-  worker = lua_touserdata(lua, 1);
-  lua_pop(lua, 1);
+  lua_getfield(L, LUA_REGISTRYINDEX, "htt_worker");
+  worker = lua_touserdata(L, 1);
+  lua_pop(L, 1);
 
-  lua_getfield(lua, LUA_REGISTRYINDEX, "htt_parent");
-  parent = lua_touserdata(lua, 1);
-  lua_pop(lua, 1);
+  lua_getfield(L, LUA_REGISTRYINDEX, "htt_parent");
+  parent = lua_touserdata(L, 1);
+  lua_pop(L, 1);
 
   apr_pool_create(&ptmp, worker->heartbeat);
 
@@ -197,7 +201,7 @@ static int lua_interpret(lua_State *lua) {
 
   line = apr_strtok(buffer, "\n", &last);
   while (line) {
-    while (*line == ' ') ++line;
+    while (*line == ' ') { ++line ; }
     if (*line != '\0') {
       apr_table_add(lines, "lua inline", line);
     }
@@ -208,7 +212,7 @@ static int lua_interpret(lua_State *lua) {
   call->interpret = parent->interpret;
 
   if ((status = call->interpret(call, worker, ptmp)) != APR_SUCCESS) {
-    luaL_error(lua, "Error: %s(%d)", my_status_str(ptmp, status), status);
+    luaL_error(L, "Error: %s(%d)", my_status_str(ptmp, status), status);
     return 1;
   }
 
@@ -216,38 +220,168 @@ static int lua_interpret(lua_State *lua) {
 }
 
 /**
- * This is test function
- * @lua_arg string IN variable name
- * @lua_return variable value
- * @param lua IN lua state
+ * Get variable from httest
+ * @param L IN lua state
  * @return 0
  */
-static int lua_getvar(lua_State *lua) {
+static int luam_getvar(lua_State *L) {
   worker_t *worker;
   const char *val;
 
-  const char *var = lua_tostring(lua, -1);
+  const char *var = lua_tostring(L, -1);
 
-  lua_pop(lua, 1);
+  lua_pop(L, 1);
 
-  lua_getfield(lua, LUA_REGISTRYINDEX, "htt_worker");
-  worker = lua_touserdata(lua, 1);
+  lua_getfield(L, LUA_REGISTRYINDEX, "htt_worker");
+  worker = lua_touserdata(L, 1);
  
   if ((val = worker_var_get(worker, var))) {
-    lua_pushstring(lua, val);
+    lua_pushstring(L, val);
     return 1;
   }
 
   return 0;
 }
 
+/**
+ * Get transport object.
+ * @param L IN lua state
+ * @return 0
+ */
+static int luam_transport_get(lua_State *L) {
+  worker_t *worker;
+
+  lua_getfield(L, LUA_REGISTRYINDEX, "htt_worker");
+  worker = lua_touserdata(L, 1);
+
+  if (!worker->socket || !worker->socket->transport) {
+    lua_pushnil(L);
+    return 1;
+  }
+
+  lua_pushlightuserdata(L, worker->socket->transport);
+    
+  luaL_getmetatable(L, "htt.transport");
+  lua_setmetatable(L, -2);
+    
+  return 1;
+}
+
+/**
+ * Get transport object.
+ * @param L IN lua state
+ * @return 0
+ */
+static transport_t *lua_checktransport (lua_State *L) {
+  void *ud = luaL_checkudata(L, 1, "htt.transport");
+  luaL_argcheck(L, ud != NULL, 1, "`transport' expected");
+  return (transport_t *)ud;
+}
+
+/**
+ * Transport read method
+ * @param L IN lua state
+ * @return 0
+ */
+static int luam_transport_read(lua_State *L) {
+  if (lua_isnumber(L, -1)) {
+    apr_status_t status;
+    apr_pool_t *pool;
+    apr_size_t bytes;
+    transport_t *transport;
+    char *buffer;
+
+    bytes = lua_tointeger(L, -1);
+    transport = lua_checktransport(L);
+    apr_pool_create(&pool, NULL);
+    buffer = apr_pcalloc(pool, bytes);
+    if ((status = transport_read(transport, buffer, &bytes)) != APR_SUCCESS) {
+      lua_pushnil(L);
+      apr_pool_destroy(pool);
+      return 1;
+    }
+    lua_pushlstring(L, buffer, bytes);
+    apr_pool_destroy(pool);
+    return 1;
+  }
+  else {
+    luaL_error(L, "Expect number of bytes");
+    return 1;
+  }
+}
+
+/**
+ * Transport write method
+ * @param L IN lua state
+ * @return 0
+ */
+static int luam_transport_write(lua_State *L) {
+  if (lua_isstring(L, -1)) {
+    apr_status_t status;
+    apr_size_t bytes;
+    const char *buffer = lua_tolstring(L, -1, &bytes);
+    transport_t *transport = lua_checktransport(L);
+    if ((status = transport_write(transport, buffer, bytes)) != APR_SUCCESS) {
+      luaL_error(L, "Could not write %d bytes", bytes);
+      return 1;
+    }
+  }
+  return 0;
+}
+
+/**
+ * Transport set timeout method
+ * @param L IN lua state
+ * @return 0
+ */
+static int luam_transport_set_timeout(lua_State *L) {
+  if (lua_isnumber(L, -1)) {
+    apr_status_t status;
+    apr_interval_time_t tmo = lua_tointeger(L, -1);
+    transport_t *transport = lua_checktransport(L);
+    if ((status = transport_set_timeout(transport, tmo * 1000)) != APR_SUCCESS) {
+      luaL_error(L, "Could not timeout %d ms", tmo);
+      return 1;
+    }
+  }
+  return 0;
+}
+
+/**
+ * Transport get timeout method
+ * @param L IN lua state
+ * @return 1
+ */
+static int luam_transport_get_timeout(lua_State *L) {
+  apr_status_t status;
+  apr_interval_time_t tmo;
+  transport_t *transport = lua_checktransport(L);
+  if ((status = transport_get_timeout(transport, &tmo)) != APR_SUCCESS) {
+    luaL_error(L, "Could not get timeout");
+    return 1;
+  }
+  else {
+    lua_pushinteger(L, tmo/1000);
+    return 1;
+  }
+}
+
 /**
  * Set of htt commands for lua
  */
-static const struct luaL_Reg httlib[] = {
-  {"version", lua_version},
-  {"interpret", lua_interpret},
-  {"getvar", lua_getvar},
+static const struct luaL_Reg htt_lib_f[] = {
+  {"version", luam_version},
+  {"interpret", luam_interpret},
+  {"getVar", luam_getvar},
+  {"getTransport", luam_transport_get},
+  {NULL, NULL}
+};
+
+static const struct luaL_Reg htt_transport_m[] = {
+  {"read", luam_transport_read},
+  {"write", luam_transport_write},
+  {"setTimeout", luam_transport_set_timeout},
+  {"getTimeout", luam_transport_get_timeout},
   {NULL, NULL}
 };
 
@@ -260,61 +394,63 @@ static const struct luaL_Reg httlib[] = {
  */
 static apr_status_t block_lua_interpreter(worker_t *worker, worker_t *parent, 
                                           apr_pool_t *ptmp) {
-	int i;
+  int failed;
+  int i;
   apr_table_entry_t *e; 
   lua_reader_t *reader;
 
   lua_wconf_t *config = lua_get_worker_config(worker);
-  lua_State *lua = lua_open();
+  lua_State *L = luaL_newstate();
 
-  luaL_openlibs(lua);
+  luaL_openlibs(L);
   e = (apr_table_entry_t *) apr_table_elts(config->params)->elts;
   for (i = 1; i < apr_table_elts(config->params)->nelts; i++) {
     const char *val = NULL;
     char *param = store_get_copy(worker->params, ptmp, e[i].key);
-    if (strncmp(param, "VAR(", 4) == 0) {
-      char *var = param + 4;
-      apr_size_t len = strlen(var);
-      if (len > 0) {
-        var[len-1] = 0;
-      }
-      val = store_get(worker->vars, var);
-      if (!val) {
-        val = store_get(worker->locals, var);
-      }
-      if (!val) {
-        val = param;
-      }
-      }
-      else {
-        val = param;
-      }
-    lua_pushstring(lua, val);
-    lua_setglobal(lua, e[i].key);
+    val = worker_get_value_from_param(worker, param, ptmp);
+    lua_pushstring(L, val);
+    lua_setglobal(L, e[i].key);
   }
-  lua_pushlightuserdata(lua, worker);
-  lua_setfield(lua, LUA_REGISTRYINDEX, "htt_worker");
-  lua_pushlightuserdata(lua, parent);
-  lua_setfield(lua, LUA_REGISTRYINDEX, "htt_parent");
-  luaL_register(lua, "htt", httlib);
+  luaopen_crypto(L);
+  lua_pushlightuserdata(L, parent);
+  lua_setfield(L, LUA_REGISTRYINDEX, "htt_parent");
+  lua_pushlightuserdata(L, worker);
+  lua_setfield(L, LUA_REGISTRYINDEX, "htt_worker");
+  luaL_newmetatable(L, "htt.transport");
+  lua_pushvalue(L, -1);  /* pushes the metatable */
+  lua_setfield(L, -2, "__index");  /* metatable.__index = metatable */
+  luaL_register(L, NULL, htt_transport_m);
+  luaL_register(L, "htt", htt_lib_f);
+  lua_pop(L, -1);
+  lua_pop(L, -1);
   reader = lua_new_lua_reader(worker, ptmp);
-  if (lua_load(lua, lua_get_line, reader, "@client") != 0 ||
-      lua_pcall(lua, 0, LUA_MULTRET, 0) != 0) {
-    const char *msg = lua_tostring(lua, -1);
-    if (msg == NULL) msg = "(error object is not a string)";
+#if ( LUA_VERSION_NUM == 501 )		
+  failed = (lua_load(L, lua_get_line, reader, "@client") != 0 || 
+            lua_pcall(L, 0, LUA_MULTRET, 0) != 0);
+#elif ( LUA_VERSION_NUM  == 502 )
+  failed = (lua_load(L, lua_get_line, reader, "@client", NULL) != 0 || 
+            lua_pcall(L, 0, LUA_MULTRET, 0) != 0);
+#else
+  #error this lua version is not supported
+#endif
+  if (failed) {
+    const char *msg = lua_tostring(L, -1);
+    if (msg == NULL) {
+      msg = "(error object is not a string)";
+    }
     worker_log_error(worker, "Lua error: %s", msg);
-    lua_pop(lua, 1);
+    lua_pop(L, 1);
     return APR_EGENERAL;
   }
   e = (apr_table_entry_t *) apr_table_elts(config->retvars)->elts;
   for (i = 0; i < apr_table_elts(config->retvars)->nelts; i++) {
     worker_log(worker, LOG_DEBUG, "param: %s; val: %s", e[i].key, e[i].val);
-    if (lua_isstring(lua, i + 2)) {
-      store_set(worker->vars, store_get(worker->retvars, e[i].key), lua_tostring(lua, i + 2));
+    if (lua_isstring(L, i + 1)) {
+      store_set(worker->vars, store_get(worker->retvars, e[i].key), lua_tostring(L, i + 1));
     }
   }
   
-  lua_close(lua);
+  lua_close(L);
 
   return APR_SUCCESS;
 }
@@ -417,6 +553,7 @@ static apr_status_t lua_block_end(global_t *global) {
  * Module
  ***********************************************************************/
 apr_status_t lua_module_init(global_t *global) {
+  module_command_new(global, "LUA", "_MODULE", "", "", NULL);
   htt_hook_block_start(lua_block_start, NULL, NULL, 0);
   htt_hook_read_line(lua_read_line, NULL, NULL, 0);
   htt_hook_block_end(lua_block_end, NULL, NULL, 0);
