@@ -1,8 +1,9 @@
-/* Licensed to the Apache Software Foundation (ASF) under one or more
- * contributor license agreements. 
- * The ASF licenses this file to You under the Apache License, Version 2.0
- * (the "License"); you may not use this file except in compliance with
- * the License.  You may obtain a copy of the License at
+/**
+ * Copyright 2010 Christian Liesch
+ *
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
  *
  *     http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -26,21 +27,12 @@
  ***********************************************************************/
 #include <openssl/ssl.h>
 #include <openssl/pkcs12.h>
-
-/* on windows the inclusion of windows.h/wincrypt.h causes
- * X509_NAME and a few more to be defined; found no other
- * way than to undef manually before inclusion of engine.h;
- * somehow the same undef in ossl_typ.h is not enough...
- */
-#ifdef OPENSSL_SYS_WIN32
-#undef X509_NAME
-#endif
+#include "module.h"
 
 #ifndef OPENSSL_NO_ENGINE
 #include <openssl/engine.h>
 #endif
 
-#include "module.h"
 #include "ssl.h"
 
 /************************************************************************
@@ -58,9 +50,15 @@ typedef struct ssl_config_s {
   const char *certfile;
   const char *keyfile;
   const char *cafile;
+  const char *set_ca;
+  const char *cipher_suite;
 #define SSL_CONFIG_FLAGS_NONE 0
 #define SSL_CONFIG_FLAGS_CERT_SET 1
+#define SSL_CONFIG_FLAGS_TRACE 2
   int flags;
+  apr_pool_t *msg_pool;
+  apr_table_t *msgs;
+  worker_t *msg_worker;
 } ssl_config_t;
 
 typedef struct ssl_socket_config_s {
@@ -118,11 +116,13 @@ static ssl_config_t *ssl_get_worker_config(worker_t *worker) {
  * @return socket config
  */
 static ssl_socket_config_t *ssl_get_socket_config(worker_t *worker) {
+  ssl_socket_config_t *config;
+
   if (!worker || !worker->socket) {
     return NULL;
   }
 
-  ssl_socket_config_t *config = module_get_config(worker->socket->config, ssl_module);
+  config = module_get_config(worker->socket->config, ssl_module);
   if (config == NULL) {
     config = apr_pcalloc(worker->pbody, sizeof(*config));
     module_set_config(worker->socket->config, apr_pstrdup(worker->pbody, ssl_module), config);
@@ -205,6 +205,378 @@ static apr_status_t worker_ssl_ctx_p12(worker_t * worker, const char *infile,
   return APR_SUCCESS;
 }
 
+/**
+ * tls ext call back for debugging
+ * @param ssl IN ssl instance
+ * @param client_server IN
+ * @param type IN
+ * @param data IN 
+ * @param len IN 
+ * @param arg IN void pointer to worker
+ */
+static void ssl_tlsext_trace(SSL *s, int client_server, int type, unsigned char *data, int len, void *arg) {
+  worker_t *worker = arg;
+  char *extname;
+  char *entry;
+  ssl_config_t *config = ssl_get_worker_config(worker);
+
+  switch(type) {
+    case TLSEXT_TYPE_server_name:
+      extname = "server name";
+      break;
+
+    case TLSEXT_TYPE_max_fragment_length:
+      extname = "max fragment length";
+      break;
+
+    case TLSEXT_TYPE_client_certificate_url:
+      extname = "client certificate URL";
+      break;
+
+    case TLSEXT_TYPE_trusted_ca_keys:
+      extname = "trusted CA keys";
+      break;
+
+    case TLSEXT_TYPE_truncated_hmac:
+      extname = "truncated HMAC";
+      break;
+
+    case TLSEXT_TYPE_status_request:
+      extname = "status request";
+      break;
+
+    case TLSEXT_TYPE_elliptic_curves:
+      extname = "elliptic curves";
+      break;
+
+    case TLSEXT_TYPE_ec_point_formats:
+      extname = "EC point formats";
+      break;
+
+    case TLSEXT_TYPE_session_ticket:
+      extname = "server ticket";
+      break;
+
+    case TLSEXT_TYPE_renegotiate:
+      extname = "renegotiate";
+      break;
+
+#ifdef TLSEXT_TYPE_opaque_prf_input
+    case TLSEXT_TYPE_opaque_prf_input:
+      extname = "opaque PRF input";
+      break;
+#endif
+
+    default:
+      extname = "unknown";
+      break;
+
+  }
+
+  entry = apr_psprintf(config->msg_pool, "+TLS %s extension, %s, id=%d", 
+                       client_server ? "server": "client",
+                       extname, type);
+  apr_table_addn(config->msgs, apr_psprintf(config->msg_pool, "TRUE"), entry);
+  worker_log(worker, LOG_INFO, "%s", entry);
+
+}
+
+/**
+ * Message call back for debugging
+ * @param write_p IN write/read
+ * @param version IN SSL version
+ * @param content_type IN SSL message content type
+ * @param buf IN SSL message
+ * @param len IN SSL message length
+ * @param ssl IN ssl instance
+ * @param arg IN void pointer to worker
+ */
+static void ssl_message_trace(int write_p, int version, int content_type, const void *buf, size_t len, SSL *ssl, void *arg)
+{
+  worker_t *worker = arg;
+  char *entry;
+  const char *str_write_p, *str_version, *str_content_type = "", *str_details1 = "", *str_details2= "";
+  ssl_config_t *config = ssl_get_worker_config(worker);
+
+  str_write_p = write_p ? ">" : "<";
+
+  switch (version) {
+    case SSL2_VERSION:
+      str_version = "SSL 2.0";
+      break;
+    case SSL3_VERSION:
+      str_version = "SSL 3.0";
+      break;
+    case TLS1_VERSION:
+      str_version = "TLS 1.0";
+      break;
+#if (OPENSSL_VERSION_NUMBER >= 0x1000100fL)
+    case TLS1_1_VERSION:
+      str_version = "TLS 1.1";
+      break;
+    case TLS1_2_VERSION:
+      str_version = "TLS 1.2";
+      break;
+#endif
+    case DTLS1_VERSION:
+      str_version = "DTLS 1.0";
+      break;
+    case DTLS1_BAD_VER:
+      str_version = "DTLS 1.0 (bad)";
+      break;
+    default:
+      str_version = "???";
+  }
+
+  if (version == SSL2_VERSION) {
+    str_details1 = "???";
+
+    if (len > 0) {
+      switch (((const unsigned char*)buf)[0]) {
+        case 0:
+          str_details1 = ", ERROR:";
+          str_details2 = " ???";
+          if (len >= 3) {
+            unsigned err = (((const unsigned char*)buf)[1]<<8) + ((const unsigned char*)buf)[2];
+
+            switch (err) {
+              case 0x0001:
+                str_details2 = " NO-CIPHER-ERROR";
+                break;
+              case 0x0002:
+                str_details2 = " NO-CERTIFICATE-ERROR";
+                break;
+              case 0x0004:
+                str_details2 = " BAD-CERTIFICATE-ERROR";
+                break;
+              case 0x0006:
+                str_details2 = " UNSUPPORTED-CERTIFICATE-TYPE-ERROR";
+                break;
+            }
+          }
+
+          break;
+        case 1:
+          str_details1 = ", CLIENT-HELLO";
+          break;
+        case 2:
+          str_details1 = ", CLIENT-MASTER-KEY";
+          break;
+        case 3:
+          str_details1 = ", CLIENT-FINISHED";
+          break;
+        case 4:
+          str_details1 = ", SERVER-HELLO";
+          break;
+        case 5:
+          str_details1 = ", SERVER-VERIFY";
+          break;
+        case 6:
+          str_details1 = ", SERVER-FINISHED";
+          break;
+        case 7:
+          str_details1 = ", REQUEST-CERTIFICATE";
+          break;
+        case 8:
+          str_details1 = ", CLIENT-CERTIFICATE";
+          break;
+      }
+    }
+  }
+
+  if (version == SSL3_VERSION ||
+      version == TLS1_VERSION ||
+#if (OPENSSL_VERSION_NUMBER >= 0x1000100fL)
+      version == TLS1_2_VERSION ||
+      version == TLS1_1_VERSION ||
+#endif
+      version == DTLS1_VERSION ||
+      version == DTLS1_BAD_VER) {
+    switch (content_type) {
+      case 20:
+        str_content_type = "ChangeCipherSpec";
+        break;
+      case 21:
+        str_content_type = "Alert";
+        break;
+      case 22:
+        str_content_type = "Handshake";
+        break;
+    }
+
+    if (content_type == 21) {
+      str_details1 = ", ???";
+
+      if (len == 2) {
+        switch (((const unsigned char*)buf)[0])
+        {
+          case 1:
+            str_details1 = ", warning";
+            break;
+          case 2:
+            str_details1 = ", fatal";
+            break;
+        }
+
+        str_details2 = " ???";
+        switch (((const unsigned char*)buf)[1]) {
+          case 0:
+            str_details2 = " close_notify";
+            break;
+          case 10:
+            str_details2 = " unexpected_message";
+            break;
+          case 20:
+            str_details2 = " bad_record_mac";
+            break;
+          case 21:
+            str_details2 = " decryption_failed";
+            break;
+          case 22:
+            str_details2 = " record_overflow";
+            break;
+          case 30:
+            str_details2 = " decompression_failure";
+            break;
+          case 40:
+            str_details2 = " handshake_failure";
+            break;
+          case 42:
+            str_details2 = " bad_certificate";
+            break;
+          case 43:
+            str_details2 = " unsupported_certificate";
+            break;
+          case 44:
+            str_details2 = " certificate_revoked";
+            break;
+          case 45:
+            str_details2 = " certificate_expired";
+            break;
+          case 46:
+            str_details2 = " certificate_unknown";
+            break;
+          case 47:
+            str_details2 = " illegal_parameter";
+            break;
+          case 48:
+            str_details2 = " unknown_ca";
+            break;
+          case 49:
+            str_details2 = " access_denied";
+            break;
+          case 50:
+            str_details2 = " decode_error";
+            break;
+          case 51:
+            str_details2 = " decrypt_error";
+            break;
+          case 60:
+            str_details2 = " export_restriction";
+            break;
+          case 70:
+            str_details2 = " protocol_version";
+            break;
+          case 71:
+            str_details2 = " insufficient_security";
+            break;
+          case 80:
+            str_details2 = " internal_error";
+            break;
+          case 90:
+            str_details2 = " user_canceled";
+            break;
+          case 100:
+            str_details2 = " no_renegotiation";
+            break;
+          case 110:
+            str_details2 = " unsupported_extension";
+            break;
+          case 111:
+            str_details2 = " certificate_unobtainable";
+            break;
+          case 112:
+            str_details2 = " unrecognized_name";
+            break;
+          case 113:
+            str_details2 = " bad_certificate_status_response";
+            break;
+          case 114:
+            str_details2 = " bad_certificate_hash_value";
+            break;
+          case 115:
+            str_details2 = " unknown_psk_identity";
+            break;
+        }
+      }
+    }
+
+    if (content_type == 22) {
+      str_details1 = "???";
+
+      if (len > 0) {
+        switch (((const unsigned char*)buf)[0]) {
+          case 0:
+            str_details1 = ", HelloRequest";
+            break;
+          case 1:
+            str_details1 = ", ClientHello";
+            break;
+          case 2:
+            str_details1 = ", ServerHello";
+            break;
+          case 3:
+            str_details1 = ", HelloVerifyRequest";
+            break;
+          case 11:
+            str_details1 = ", Certificate";
+            break;
+          case 12:
+            str_details1 = ", ServerKeyExchange";
+            break;
+          case 13:
+            str_details1 = ", CertificateRequest";
+            break;
+          case 14:
+            str_details1 = ", ServerHelloDone";
+            break;
+          case 15:
+            str_details1 = ", CertificateVerify";
+            break;
+          case 16:
+            str_details1 = ", ClientKeyExchange";
+            break;
+          case 20:
+            str_details1 = ", Finished";
+            break;
+        }
+      }
+    }
+
+#ifndef OPENSSL_NO_HEARTBEATS
+    if (content_type == 24) /* Heartbeat */ {
+      str_details1 = ", Heartbeat";
+
+      if (len > 0) {
+        switch (((const unsigned char*)buf)[0]) {
+          case 1:
+            str_details1 = ", HeartbeatRequest";
+            break;
+          case 2:
+            str_details1 = ", HeartbeatResponse";
+            break;
+        }
+      }
+    }
+#endif
+  }
+
+  entry = apr_psprintf(config->msg_pool, "%s%s: %s%s%s", str_write_p, 
+      str_version, str_content_type, str_details1, str_details2);
+  apr_table_addn(config->msgs, apr_psprintf(config->msg_pool, "TRUE"), entry);
+  worker_log(worker, LOG_INFO, "%s", entry);
+}
+
 /**
  * Get server ctx with loaded cert and key file
  *
@@ -221,6 +593,15 @@ static apr_status_t worker_ssl_ctx(worker_t * worker, const char *certfile,
     return APR_SUCCESS;
   }
 
+  if (certfile && !keyfile && !ca) {
+    ca = certfile;
+    certfile = NULL;
+  }
+
+  if (config->set_ca && !ca) {
+    ca = config->set_ca;
+  }
+
   /* test if there are the same cert, key ca files or no certs at all */
   if (!(
       (((!config->certfile && !certfile) || 
@@ -304,7 +685,7 @@ static apr_status_t worker_ssl_ctx(worker_t * worker, const char *certfile,
  * Get client method 
  *
  * @param worker IN thread object data
- * @param sslstr IN SSL|SSL2|SSL3|TLS1
+ * @param sslstr IN SSL|SSL2|SSL3|TLS1|DTLS1
  *
  * @return APR_SUCCESS or APR_ECONNABORTED
  */
@@ -330,6 +711,20 @@ static int worker_set_client_method(worker_t * worker, const char *sslstr) {
     is_ssl = 1;
     config->meth = TLSv1_client_method();
   }
+#if (OPENSSL_VERSION_NUMBER >= 0x1000100fL)
+  else if (strcasecmp(sslstr, "TLS1.1") == 0) {
+    is_ssl = 1;
+    config->meth = TLSv1_1_client_method();
+  }
+  else if (strcasecmp(sslstr, "TLS1.2") == 0) {
+    is_ssl = 1;
+    config->meth = TLSv1_2_client_method();
+  }
+#endif
+  else if (strcasecmp(sslstr, "DTLS1") == 0) {
+    is_ssl = 1;
+    config->meth = DTLSv1_client_method();
+  }
   return is_ssl;
 }
 
@@ -337,7 +732,7 @@ static int worker_set_client_method(worker_t * worker, const char *sslstr) {
  * Get server method 
  *
  * @param worker IN thread object data
- * @param sslstr IN SSL|SSL2|SSL3|TLS1
+ * @param sslstr IN SSL|SSL2|SSL3|TLS1|DTLS1
  *
  * @return APR_SUCCESS or APR_ECONNABORTED
  */
@@ -363,9 +758,60 @@ static int worker_set_server_method(worker_t * worker, const char *sslstr) {
     is_ssl = 1;
     config->meth = TLSv1_server_method();
   }
+#if (OPENSSL_VERSION_NUMBER >= 0x1000100fL)
+  else if (strcasecmp(sslstr, "TLS1.1") == 0) {
+    is_ssl = 1;
+    config->meth = TLSv1_1_server_method();
+  }
+  else if (strcasecmp(sslstr, "TLS1.2") == 0) {
+    is_ssl = 1;
+    config->meth = TLSv1_2_server_method();
+  }
+#endif
+  else if (strcasecmp(sslstr, "DTLS1") == 0) {
+    is_ssl = 1;
+    config->meth = DTLSv1_server_method();
+  }
   return is_ssl;
 }
 
+/**
+ * create new ssl instance
+ *
+ * @param worker IN callee
+ * @return apr status
+ */
+static apr_status_t ssl_new_instance(worker_t *worker) {
+  ssl_config_t *config = ssl_get_worker_config(worker);
+  ssl_socket_config_t *sconfig = ssl_get_socket_config(worker);
+
+  if (config->msg_worker) {
+    worker_destroy(config->msg_worker);
+  }
+  worker_new(&config->msg_worker, NULL, worker->prefix, worker->global, worker->interpret); 
+  config->msg_worker->config = worker->config;
+
+  if ((sconfig->ssl = SSL_new(config->ssl_ctx)) == NULL) {
+    worker_log(worker, LOG_ERR, "SSL_new failed.");
+    return APR_ECONNREFUSED;
+  }
+  SSL_set_ssl_method(sconfig->ssl, config->meth);
+  if (config->flags & SSL_CONFIG_FLAGS_TRACE) {
+#ifndef OPENSSL_NO_TLSEXT
+    SSL_set_tlsext_debug_callback(sconfig->ssl, ssl_tlsext_trace);
+    SSL_set_tlsext_debug_arg(sconfig->ssl, config->msg_worker);
+#endif
+    SSL_set_msg_callback(sconfig->ssl, ssl_message_trace);
+    SSL_set_msg_callback_arg(sconfig->ssl, config->msg_worker);
+  }
+  if (config->cipher_suite != NULL) {
+    if (SSL_set_cipher_list(sconfig->ssl, config->cipher_suite) == 0) {
+      return APR_EINVAL;
+    }
+  }
+  return APR_SUCCESS;
+}
+
 /**
  * Do a ssl accept
  *
@@ -376,7 +822,6 @@ static int worker_set_server_method(worker_t * worker, const char *sslstr) {
 static apr_status_t worker_ssl_accept(worker_t * worker) {
   apr_status_t status;
   char *error;
-  ssl_config_t *config = ssl_get_worker_config(worker);
   ssl_socket_config_t *sconfig = ssl_get_socket_config(worker);
 
   if (worker->socket->is_ssl) {
@@ -384,11 +829,16 @@ static apr_status_t worker_ssl_accept(worker_t * worker) {
       BIO *bio;
       apr_os_sock_t fd;
 
-      if ((sconfig->ssl = SSL_new(config->ssl_ctx)) == NULL) {
-	worker_log(worker, LOG_ERR, "SSL_new failed.");
-	status = APR_ECONNREFUSED;
+      if ((status = ssl_new_instance(worker)) != APR_SUCCESS) {
+        return status;
       }
-      SSL_set_ssl_method(sconfig->ssl, config->meth);
+
+      /**
+       * openssl 1.0.1b want this ...
+       * Have to understand this first, before I do add it
+      SSL_set_session_id_context(sconfig->ssl, "foobar", strlen("foobar"));
+       */
+
       ssl_rand_seed();
       apr_os_sock_get(&fd, worker->socket->socket);
       bio = BIO_new_socket(fd, BIO_NOCLOSE);
@@ -416,7 +866,7 @@ static apr_status_t worker_ssl_accept(worker_t * worker) {
  * @param desc OUT os socket descriptor
  * @return APR_ENOENT
  */
-apr_status_t ssl_transport_os_desc_get(void *data, int *desc) {
+static apr_status_t ssl_transport_os_desc_get(void *data, int *desc) {
   return APR_ENOENT;
 }
 
@@ -427,7 +877,7 @@ apr_status_t ssl_transport_os_desc_get(void *data, int *desc) {
  * @param t IN timeout 
  * @return APR_ENOENT
  */
-apr_status_t ssl_transport_set_timeout(void *data, apr_interval_time_t t) {
+static apr_status_t ssl_transport_set_timeout(void *data, apr_interval_time_t t) {
   ssl_transport_t *ssl_transport = data;
 
   ssl_transport->tmo = t;
@@ -441,7 +891,7 @@ apr_status_t ssl_transport_set_timeout(void *data, apr_interval_time_t t) {
  * @param t OUT timeout 
  * @return APR_ENOENT
  */
-apr_status_t ssl_transport_get_timeout(void *data, apr_interval_time_t *t) {
+static apr_status_t ssl_transport_get_timeout(void *data, apr_interval_time_t *t) {
   ssl_transport_t *ssl_transport = data;
 
   return transport_get_timeout(ssl_transport->tcp_transport, t);
@@ -455,7 +905,7 @@ apr_status_t ssl_transport_get_timeout(void *data, apr_interval_time_t *t) {
  * @param size INOUT buffer len
  * @return apr status
  */
-apr_status_t ssl_transport_read(void *data, char *buf, apr_size_t *size) {
+static apr_status_t ssl_transport_read(void *data, char *buf, apr_size_t *size) {
   ssl_transport_t *ssl_transport = data;
   apr_status_t status;
   apr_time_t start = apr_time_now();
@@ -500,7 +950,7 @@ tryagain:
  * @param size INOUT buffer len
  * @return apr status
  */
-apr_status_t ssl_transport_write(void *data, char *buf, apr_size_t size) {
+static apr_status_t ssl_transport_write(void *data, const char *buf, apr_size_t size) {
   ssl_transport_t *ssl_transport = data;
   apr_size_t e_ssl;
 
@@ -534,7 +984,6 @@ static apr_status_t block_SSL_CONNECT(worker_t * worker, worker_t *parent, apr_p
   int is_ssl;
   BIO *bio;
   apr_os_sock_t fd;
-  ssl_config_t *config = ssl_get_worker_config(worker);
   ssl_socket_config_t *sconfig = ssl_get_socket_config(worker);
 
   sslstr = store_get(worker->params, "1");
@@ -566,11 +1015,10 @@ static apr_status_t block_SSL_CONNECT(worker_t * worker, worker_t *parent, apr_p
 	return status;
       }
 
-      if ((sconfig->ssl = SSL_new(config->ssl_ctx)) == NULL) {
-				worker_log(worker, LOG_ERR, "SSL_new failed.");
-				return APR_ECONNREFUSED;
+      if ((status = ssl_new_instance(worker)) != APR_SUCCESS) {
+        return status;
       }
-      SSL_set_ssl_method(sconfig->ssl, config->meth);
+
       ssl_rand_seed();
       apr_os_sock_get(&fd, worker->socket->socket);
       bio = BIO_new_socket(fd, BIO_NOCLOSE);
@@ -839,19 +1287,25 @@ static apr_status_t block_SSL_RENEG_CERT(worker_t * worker, worker_t *parent, ap
   }
 
   if (worker->flags & FLAGS_SERVER) {
+    /* if we are server request the peer cert */
     if (copy && strcasecmp(copy, "verify") == 0) {
-      /* if we are server request the peer cert */
       if (worker->log_mode >= LOG_DEBUG) {
-	SSL_set_verify(sconfig->ssl,
-		       SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
-		       debug_verify_callback);
+        SSL_set_verify(sconfig->ssl,
+                       SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
+                       debug_verify_callback);
       }
       else {
-	SSL_set_verify(sconfig->ssl,
-		       SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
-		       NULL);
+        SSL_set_verify(sconfig->ssl,
+                       SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
+                       NULL);
       }
     }
+    else {
+      SSL_set_verify(sconfig->ssl,
+                     SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
+                     skip_verify_callback);
+    }
+    SSL_set_session_id_context(sconfig->ssl, (void *)ssl_module, strlen(ssl_module));
 
     if (worker->flags & FLAGS_SSL_LEGACY) {
 #if (OPENSSL_VERSION_NUMBER >= 0x009080cf)
@@ -871,8 +1325,8 @@ static apr_status_t block_SSL_RENEG_CERT(worker_t * worker, worker_t *parent, ap
     sconfig->ssl->state=SSL_ST_ACCEPT;
     worker_ssl_handshake(worker);
 
+    config->cert = SSL_get_peer_certificate(sconfig->ssl);
     if (copy && strcasecmp(copy, "verify") == 0) {
-      config->cert = SSL_get_peer_certificate(sconfig->ssl);
       if (!config->cert) {
 	worker_log(worker, LOG_ERR, "No peer certificate");
 	return APR_EACCES;
@@ -880,13 +1334,13 @@ static apr_status_t block_SSL_RENEG_CERT(worker_t * worker, worker_t *parent, ap
     }
   }
   else {
-    if (copy && strcasecmp(copy, "verify") == 0) {
-      config->cert = SSL_get_peer_certificate(sconfig->ssl);
-      if (!config->cert) {
-	worker_log(worker, LOG_ERR, "No peer certificate");
-	return APR_EACCES;
-      }
+    config->cert = SSL_get_peer_certificate(sconfig->ssl);
+    if (!config->cert) {
+      worker_log(worker, LOG_ERR, "No peer certificate");
+      return APR_EACCES;
+    }
 
+    if (copy && strcasecmp(copy, "verify") == 0) {
       if((rc = SSL_get_verify_result(sconfig->ssl)) != X509_V_OK) {
 	worker_log(worker, LOG_ERR, "SSL peer verify failed: %s(%d)",
 	X509_verify_cert_error_string(rc), rc);
@@ -961,6 +1415,25 @@ static apr_status_t block_SSL_SET_ENGINE(worker_t * worker, worker_t *parent, ap
   return APR_ENOTIMPL;
 }
 
+/**
+ * SSL_CIPHER_SUITE command
+ *
+ * See http://www.openssl.org/docs/apps/ciphers.html
+ *
+ * @param worker IN thread data object
+ * @param data IN
+ *
+ * @return APR_SUCCESS or APR_EINVAL
+ */
+static apr_status_t block_SSL_CIPHER_SUITE(worker_t * worker,
+                                                     worker_t *parent, apr_pool_t *ptmp) {
+  ssl_config_t *config = ssl_get_worker_config(worker);
+  const char *val = store_get(worker->params, "1");
+  char *copy = apr_pstrdup(worker->pbody, val);
+  config->cipher_suite = copy;
+  return APR_SUCCESS;
+}
+
 /**
  * SSL_LEGACY command
  *
@@ -1070,8 +1543,8 @@ static apr_status_t block_SSL_SET_CERT(worker_t * worker, worker_t *parent, apr_
   }
 
   /* else set cert */
-  if (!config || !config->cert || !config->pkey) {
-    worker_log_error(worker, "No cert and key to use, get a cert and key with _SSL:LOAD_KEY of _SSL:LOAD_CERT");
+  if (!config || !config->cert) {
+    worker_log_error(worker, "No cert to use, get a cert with _SSL:LOAD_CERT");
     return APR_EINVAL;
   }
 
@@ -1082,6 +1555,26 @@ static apr_status_t block_SSL_SET_CERT(worker_t * worker, worker_t *parent, apr_
   return APR_SUCCESS;
 }
 
+/**
+ * SSL_SET_CA command
+ *
+ * @param worker IN thread data object
+ * @param data IN ssl variable and a variable name 
+ *
+ * @return APR_SUCCESS
+ */
+static apr_status_t block_SSL_SET_CA(worker_t * worker, worker_t *parent, apr_pool_t *ptmp) {
+  ssl_config_t *config = ssl_get_worker_config(worker);
+
+  /* check if we have parameters */
+  if (store_get_size(worker->params)) {
+    config->set_ca = store_get(worker->params, "1");
+    return  APR_SUCCESS;
+  }
+
+  return APR_SUCCESS;
+}
+
 /**
  * SSL_SECURE_RENEG_SUPPORTED command
  *
@@ -1105,6 +1598,25 @@ static apr_status_t block_SSL_SECURE_RENEG_SUPPORTED(worker_t * worker,
 #endif
 }
 
+/**
+ * SSL_TRACE_START command
+ *
+ * @param worker IN thread data object
+ * @param data IN 
+ *
+ * @return APR_SUCCESS or apr error code
+ */
+static apr_status_t block_SSL_TRACE(worker_t * worker, worker_t *parent, 
+                                    apr_pool_t *ptmp) {
+  apr_pool_t *pool;
+  ssl_config_t *config = ssl_get_worker_config(worker);
+  config->flags |= SSL_CONFIG_FLAGS_TRACE;
+  apr_pool_create(&pool, NULL);
+  config->msg_pool = pool;
+  config->msgs = apr_table_make(pool, 5);
+  return APR_SUCCESS;
+}
+
 /**
  * clone worker
  *
@@ -1228,7 +1740,6 @@ static apr_status_t ssl_server_port_args(worker_t *worker, char *portinfo,
  */
 static apr_status_t ssl_hook_connect(worker_t *worker) {
   apr_status_t status;
-  ssl_config_t *config = ssl_get_worker_config(worker);
   ssl_socket_config_t *sconfig = ssl_get_socket_config(worker);
 
   if (worker->socket->is_ssl) {
@@ -1237,11 +1748,10 @@ static apr_status_t ssl_hook_connect(worker_t *worker) {
     BIO *bio;
     apr_os_sock_t fd;
 
-    if ((sconfig->ssl = SSL_new(config->ssl_ctx)) == NULL) {
-      worker_log(worker, LOG_ERR, "SSL_new failed.");
-      return APR_EGENERAL;
+    if ((status = ssl_new_instance(worker)) != APR_SUCCESS) {
+      return status;
     }
-    SSL_set_ssl_method(sconfig->ssl, config->meth);
+
     ssl_rand_seed();
     apr_os_sock_get(&fd, worker->socket->socket);
     bio = BIO_new_socket(fd, BIO_NOCLOSE);
@@ -1281,6 +1791,8 @@ static apr_status_t ssl_hook_accept(worker_t *worker, char *data) {
 
   if (worker->socket->is_ssl) {
     char *last;
+    transport_t *transport;
+    ssl_transport_t *ssl_transport;
     char *cert = NULL;
     char *key = NULL;
     char *ca = NULL;
@@ -1304,8 +1816,6 @@ static apr_status_t ssl_hook_accept(worker_t *worker, char *data) {
       return status;
     }
 
-    transport_t *transport;
-    ssl_transport_t *ssl_transport;
     if ((status = worker_ssl_accept(worker)) != APR_SUCCESS) {
       return status;
     }
@@ -1358,6 +1868,34 @@ static apr_status_t ssl_hook_close(worker_t *worker, char *info,
   return APR_SUCCESS;
 }
 
+/**
+ * expect/grep/match SSL handshake
+ *
+ * @param worker IN
+ *
+ * @return APR_SUCCESS or apr error
+ */
+static apr_status_t ssl_hook_read_pre_headers(worker_t *worker) {
+  ssl_config_t *config = ssl_get_worker_config(worker);
+  if (config->msgs) {
+    int i;
+    apr_table_entry_t *e = (apr_table_entry_t *) apr_table_elts(config->msgs)->elts;
+    for (i = 0; i < apr_table_elts(config->msgs)->nelts; i++) {
+      worker_match(worker, worker->match.dot, e[i].val, strlen(e[i].val));
+      worker_match(worker, worker->match.headers, e[i].val, strlen(e[i].val));
+      worker_match(worker, worker->grep.dot, e[i].val, strlen(e[i].val));
+      worker_match(worker, worker->grep.headers, e[i].val, strlen(e[i].val));
+      worker_expect(worker, worker->expect.dot, e[i].val, strlen(e[i].val));
+      worker_expect(worker, worker->expect.headers, e[i].val, strlen(e[i].val));
+    }
+    apr_table_clear(config->msgs);
+    apr_pool_destroy(config->msg_pool);
+    apr_pool_create(&config->msg_pool, NULL);
+    config->msgs = apr_table_make(config->msg_pool, 5);
+  }
+  return APR_SUCCESS;
+}
+
 /************************************************************************
  * Module 
  ***********************************************************************/
@@ -1378,14 +1916,22 @@ apr_status_t ssl_module_init(global_t *global) {
   ssl_util_thread_setup(global->pool);
 
   if ((status = module_command_new(global, "SSL", "_CONNECT",
-	                           "SSL|SSL2|SSL3|TLS1 [<cert-file> <key-file>]",
+	                           "SSL|SSL2|SSL3|DTLS1|TLS1"
+#if (OPENSSL_VERSION_NUMBER >= 0x1000100fL)
+                                   "|TLS1.1|TLS1.2"
+#endif
+                                   " [<cert-file> <key-file>]",
 	                           "Needs a connected socket to establish a ssl "
 				   "connection on it.",
 	                           block_SSL_CONNECT)) != APR_SUCCESS) {
     return status;
   }
   if ((status = module_command_new(global, "SSL", "_ACCEPT",
-	                           "SSL|SSL2|SSL3|TLS1 [<cert-file> <key-file>]",
+	                           "SSL|SSL2|SSL3|DTLS|TLS1"
+#if (OPENSSL_VERSION_NUMBER >= 0x1000100fL)
+                                   "|TLS1.1|TLS1.2"
+#endif
+                                   " [<cert-file> <key-file>]",
 	                           "Needs a connected socket to accept a ssl "
 				   "connection on it.",
 	                           block_SSL_ACCEPT)) != APR_SUCCESS) {
@@ -1432,6 +1978,11 @@ apr_status_t ssl_module_init(global_t *global) {
 	                           block_SSL_SET_CERT)) != APR_SUCCESS) {
     return status;
   }
+  if ((status = module_command_new(global, "SSL", "_SET_CA", "<ca>",
+	                           "set <ca> cert",
+	                           block_SSL_SET_CA)) != APR_SUCCESS) {
+    return status;
+  }
   if ((status = module_command_new(global, "SSL", "_LOAD_KEY", "<pem-key>",
 				   "Read the given pem formated <pem-key>. Stores it for later use "
 				   "with commands like _SSL:SET_KEY",
@@ -1454,8 +2005,7 @@ apr_status_t ssl_module_init(global_t *global) {
 				   "  I_DN_<var>\n" 
 				   "  A_SIG\n" 
 				   "  A_KEY\n" 
-				   "  CERT"
-				   "Performs an SSL renegotiation.",
+				   "  CERT",
 	                           block_SSL_GET_CERT_VALUE)) != APR_SUCCESS) {
     return status;
   }
@@ -1474,6 +2024,17 @@ apr_status_t ssl_module_init(global_t *global) {
 	                           block_SSL_SECURE_RENEG_SUPPORTED)) != APR_SUCCESS) {
     return status;
   }
+  if ((status = module_command_new(global, "SSL", "_TRACE", "",
+				   "Start SSL debug session",
+	                           block_SSL_TRACE)) != APR_SUCCESS) {
+    return status;
+  }
+
+  if ((status = module_command_new(global, "SSL", "_SET_CIPHER_SUITE", "<ciphers>",
+				   "Set an opnessl cipher suite to be used",
+	                           block_SSL_CIPHER_SUITE)) != APR_SUCCESS) {
+    return status;
+  }
 
   htt_hook_worker_clone(ssl_worker_clone, NULL, NULL, 0);
   htt_hook_client_port_args(ssl_client_port_args, NULL, NULL, 0);
@@ -1481,6 +2042,7 @@ apr_status_t ssl_module_init(global_t *global) {
   htt_hook_connect(ssl_hook_connect, NULL, NULL, 0);
   htt_hook_accept(ssl_hook_accept, NULL, NULL, 0);
   htt_hook_close(ssl_hook_close, NULL, NULL, 0);
+  htt_hook_read_pre_headers(ssl_hook_read_pre_headers, NULL, NULL, 0);
   return APR_SUCCESS;
 }
 
