@@ -1,8 +1,9 @@
-/* Licensed to the Apache Software Foundation (ASF) under one or more
- * contributor license agreements. 
- * The ASF licenses this file to You under the Apache License, Version 2.0
- * (the "License"); you may not use this file except in compliance with
- * the License.  You may obtain a copy of the License at
+/**
+ * Copyright 2006 Christian Liesch
+ *
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
  *
  *     http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -164,8 +165,10 @@ struct global_s {
   apr_pool_t *pool;
   apr_hash_t *config;
   int flags;
+  const char *path;
   const char *filename;
   store_t *vars;
+  store_t *shared;
   apr_hash_t *modules;
   apr_hash_t *blocks;
   apr_table_t *files;
@@ -294,24 +297,36 @@ APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, line_get_length,
                           (worker_t *worker, line_t *line));
 APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, line_flush,
                           (worker_t *worker, line_t *line));
+APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, line_sent,
+                          (worker_t *worker, line_t *line));
 APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, client_port_args,
                           (worker_t *worker, char *portinfo, 
 			   char **new_portinfo, char *rest_of_line));
 APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, server_port_args,
                           (worker_t *worker, char *portinfo, 
 			   char **new_portinfo, char *rest_of_line));
+APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, pre_connect,
+                          (worker_t *worker));
 APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, connect,
                           (worker_t *worker));
+APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, post_connect,
+                          (worker_t *worker));
 APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, accept,
                           (worker_t *worker, char *rest_of_line));
 APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, close,
                           (worker_t *worker, char *info, char **new_info));
+APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, WAIT_begin,
+                          (worker_t *worker));
+APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, read_pre_headers,
+                          (worker_t *worker));
 APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, read_status_line,
                           (worker_t *worker, char *line));
 APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, read_header,
                           (worker_t *worker, char *line));
 APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, read_buf,
                           (worker_t *worker, char *buf, apr_size_t len));
+APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, WAIT_end,
+                          (worker_t *worker, apr_status_t status));
 APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, worker_clone,
                           (worker_t *worker, worker_t *clone));
 APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, read_line,
@@ -320,6 +335,16 @@ APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, block_start,
                           (global_t *global, char **line));
 APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, block_end,
                           (global_t *global));
+APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, client_create,
+                          (worker_t *worker, apr_thread_start_t func, apr_thread_t **new_thread));
+APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, worker_finally,
+                          (worker_t *worker));
+APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, thread_start,
+                          (global_t *global, apr_thread_t *thread));
+APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, thread_join,
+                          (global_t *global, apr_thread_t *thread));
+APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, worker_joined,
+                          (global_t *global));
 
 apr_status_t transport_register(socket_t *socket, transport_t *transport);
 apr_status_t transport_unregister(socket_t *socket, transport_t *transport);
@@ -343,6 +368,7 @@ apr_status_t command_CLOSE(command_t * self, worker_t * worker, char *data, apr_
 apr_status_t command_TIMEOUT(command_t * self, worker_t * worker, char *data, apr_pool_t *ptmp);
 apr_status_t command_MATCH(command_t * self, worker_t * worker, char *data, apr_pool_t *ptmp);
 apr_status_t command_GREP(command_t * self, worker_t * worker, char *data, apr_pool_t *ptmp);
+apr_status_t command_ASSERT(command_t * self, worker_t * worker, char *data, apr_pool_t *ptmp);
 apr_status_t command_SET(command_t * self, worker_t * worker, char *data, apr_pool_t *ptmp);
 apr_status_t command_UNSET(command_t * self, worker_t * worker, char *data, apr_pool_t *ptmp);
 apr_status_t command_DATA(command_t * self, worker_t * worker, char *data, apr_pool_t *ptmp);
@@ -399,7 +425,7 @@ apr_status_t command_DUMMY(command_t *self, worker_t *worker, char *data, apr_po
 /** helper */
 void worker_log(worker_t * self, int log_mode, char *fmt, ...); 
 void worker_log_error(worker_t * self, char *fmt, ...); 
-void worker_log_buf(worker_t * self, int log_mode, char *buf, char *prefix, 
+void worker_log_buf(worker_t * self, int log_mode, const char *buf, char *prefix, 
                     int len); 
 void worker_var_set(worker_t * worker, const char *var, const char *val); 
 const char * worker_var_get(worker_t * worker, const char *var); 
@@ -410,6 +436,7 @@ apr_status_t worker_expect(worker_t * self, apr_table_t * regexs,
                            const char *data, apr_size_t len); 
 apr_status_t worker_assert(worker_t * self, apr_status_t status); 
 apr_status_t worker_check_error(worker_t *self, apr_status_t status); 
+const char * worker_resolve_var(worker_t *worker, const char *name, apr_pool_t *ptmp); 
 char * worker_replace_vars(worker_t * worker, char *line, int *unresolved,
                            apr_pool_t *ptmp); 
 apr_status_t worker_flush(worker_t * self, apr_pool_t *ptmp);
@@ -426,5 +453,6 @@ apr_status_t worker_add_line(worker_t * self, const char *file_and_line,
 apr_status_t worker_socket_send(worker_t *self, char *buf, 
                                 apr_size_t len); 
 apr_status_t worker_to_file(worker_t * self);
+const char *worker_get_value_from_param(worker_t *worker, const char *param, apr_pool_t *ptmp); 
 
 #endif
